from .utils.rust_types import Ok, Err, Result, UnwrapOnErr, Option, Some, N, UnwrapOnNone, p, Mutex

__all__ = [
	"Ok",
	"Err",
	"Result",
	"UnwrapOnErr",
	"Option",
	"Some",
	"N",
	"UnwrapOnNone",
	"p",
	"Mutex",
]



import time, pandas as pd, numpy as np, warnings
from datetime import datetime, timedelta

"""
Each class could be called upon to, given arg:
. retrive/create an instance
  or
. retrive/create class's main value
Both will run a check for whether the object meets criterions for being of the class.
The main value is always V
"""


class ClassTemplate:
	# * If anything needs to be callable from the outside, should start with a Capital letter
	def __new__(cls, value):
		if isinstance(value, cls):
			return value
		else:
			return super().__new__(cls)

	def __init__(self, value):
		self.V = self.__class__.value(value)
		if isinstance(value, self.__class__):
			return None
		# * other needed values here

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		# * assertions here
		return value


class Market(ClassTemplate):
	def __init__(self, market):
		self.V = self.__class__.value(market)
		if isinstance(market, self.__class__):
			return None
		self.BinanceBaseUrl = "https://fapi.binance.com/fapi/v1" if self.V == "perp" else "https://api.binance.com/api/v3"

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		if value not in ["perp", "spot"]:
			raise ValueError("Valid options for market are 'perp' or 'spot'")
		return value


class USDTSymbol(ClassTemplate):
	def __init__(self, symbol):
		self.V = self.__class__.value(symbol)
		if isinstance(symbol, self.__class__):
			return None
		self.Coin = self.V[:-4].lower()

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		if not value.isupper():
			raise ValueError(f"Provided symbol is of incorrect format: {value}")
		if not value.endswith("USDT"):
			raise ValueError("USDTSymbol must be quoted against USDT")
		return value


class BinanceTf(ClassTemplate):
	def __init__(self, tf):
		self.V = self.__class__.value(tf)
		if isinstance(tf, self.__class__):
			return None
		self.Seconds = self.__class__.getSeconds(self.V)

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		if isinstance(value, int):
			value = cls.fromSeconds(value)
		# TODO: binance spot actually allows for 1s timeframes. So will have to redo this part. For now simply no warning.
		# if value not in ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M']:
		#    raise ValueError(f"Invalid tf format. Received: {value}")
		return value

	def getSeconds(tf):
		num = int(tf[:-1])
		interval = tf[-1]
		map = {"s": 1, "m": 60, "h": 60 * 60, "d": 24 * 60 * 60, "w": 7 * 24 * 60 * 60, "M": 30 * 24 * 60 * 60}
		multiplier = map.get(interval)
		seconds = num * multiplier
		return seconds

	def fromSeconds(seconds):
		atLeastTf = None
		numerator = seconds
		if numerator % 60 == 0:
			numerator /= 60
			atLeastTf = "m"
		if numerator % 60 == 0:
			numerator /= 60
			atLeastTf = "h"
		if numerator % 24 == 0:
			numerator /= 24
			atLeastTf = "d"
		if numerator % 7 == 0:
			if numerator % 30 == 0:
				numerator /= 30
				atLeastTf = "M"
			else:
				numerator /= 7
				atLeastTf = "w"
		tf = f"{int(numerator)}{atLeastTf}"
		return tf


class Timestamp(ClassTemplate):
	def __init__(self, timestamp):
		self.V = self.__class__.value(timestamp)
		if isinstance(timestamp, self.__class__):
			return None
		self.Ns = self.convertAnyToUnixNs(self.V)
		self.S = int(round(self.Ns / 1_000_000_000))
		self.Ms = int(round(self.Ns / 1_000_000))
		self.Us = int(round(self.Ns / 1_000))
		self.Datetime = datetime.fromtimestamp(self.Ns / 1_000_000_000)
		self.Isoformat = self.Datetime.isoformat()

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V  # returning the original Timestamp value, because there is no main value here
		return value

	def convertAnyToUnixNs(self, timestamp):
		unix_ns = None
		# checks if isoformat
		try:
			dt = datetime.fromisoformat(timestamp)
			unix_ns = dt.timestamp() * 1_000_000_000
		except:
			pass
		# checks if datetime
		try:
			unix_ns = timestamp.timestamp() * 1_000_000_000
		except:
			pass
		# checks if unix
		try:
			if len(str(round(timestamp))) == 10:
				unix_ns = timestamp * 1_000_000_000
			if len(str(round(timestamp))) == 13:
				unix_ns = timestamp * 1_000_000
			if len(str(round(timestamp))) == 16:
				unix_ns = timestamp * 1_000
			if len(str(round(timestamp))) == 19:
				unix_ns = timestamp
			assert 0 <= unix_ns / 1_000_000_000 <= time.time()
		except:
			pass

		assert unix_ns != None, f"Provided timestamp type isn't supported.\nProvided: {timestamp}\nSupported: [unix_s, unix_ms, unix_us, unix_ns, datetime, isoformat], where only isoformat can be string"
		return unix_ns

	# todo here should be a __minus__ (or something like that) method, returning a new self.__class__ object with all values diminished by provided number of seconds

	# todo here should be a implementation of roundinig down to a given timeframe


class ClosesDf(ClassTemplate):
	def __init__(self, closesDf):
		self.V = self.__class__.value(closesDf)
		if isinstance(closesDf, self.__class__):
			return None

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		assert isinstance(value, pd.DataFrame)
		for column in value.columns:
			assert USDTSymbol(column)
		assert isinstance(value.index[0], datetime)
		assert np.any(value.iloc[:, 0].values == 0), f"Data does not go through 0.0, so assuming the df to be not normalized. Normalize the dataframe. Currently first column is: {value.iloc[:, 0].values}"
		return value


class Klines(ClassTemplate):
	def __init__(self, klinesDf):
		self.V = self.__class__.value(klinesDf)
		if isinstance(klinesDf, self.__class__):
			return None
		self.Market: Market = None
		try:
			self.Tf: BinanceTf = self.determineTf()
		except:
			self.Tf = None
		self.Full = False  # todo
		self.Normalized = False

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		assert isinstance(value, pd.DataFrame)
		assert isinstance(value.index[0], datetime)
		mandatoryColumns = ["open", "high", "low", "close", "volume"]
		for column in mandatoryColumns:
			assert column in value.columns
		value[mandatoryColumns] = value[mandatoryColumns].apply(pd.to_numeric)
		return value

	def Normalize(self, dtTimestamp: Timestamp = None):
		if not self.Normalized:
			df = self.V
			zeroIndex = df.index[0] if dtTimestamp == None else Timestamp(dtTimestamp).datetime
			columns = ["open", "high", "low", "close", "oi", "lsr"]
			for column in columns:
				if column in df.columns:
					zeroValue = df.loc[zeroIndex, column]
					df[column] = df[column].apply(lambda x: round(np.log(x / zeroValue), 3))

	def determineTf(self):
		secondIndexTimestamp = Timestamp(self.V.index[1])
		thirdIndexTimestamp = Timestamp(self.V.index[2])
		seconds = thirdIndexTimestamp.S - secondIndexTimestamp.S
		return BinanceTf(seconds).V


class SymbolsKlines(ClassTemplate):
	def __init__(self, klinesDict):
		self.V = self.__class__.value(klinesDict)
		if isinstance(klinesDict, self.__class__):
			return None

	@classmethod
	def value(cls, value):
		if isinstance(value, cls):
			value = value.V
		assert isinstance(value, dict)
		assert len(value) >= 1
		for symbol, df in value.items():
			_, _ = USDTSymbol(symbol), Klines(df)
		return value

	def Normalize(self, dtTimestamp: Timestamp = None):
		for symbol, k in self.V.items():
			k.Normalize(dtTimestamp)

	def ToOpensDf(self) -> ClosesDf:
		self.Normalize()
		NormalizedKlinesDict = self.V
		opensDf = pd.DataFrame({symbol: k.V["close"] for symbol, k in NormalizedKlinesDict.items()})
		return opensDf



import json
import time
import pandas as pd
import os
from datetime import datetime


def dbg(*args):
    print("\n#==========================================================")
    print(args[0])
    for arg in args[1:]:
        print(f'\n{arg}')
    print("#==========================================================\n")
    alert()
    input()

# <tg>


class Telegram:
    def __init__(self):
        import telebot
        self.bot_keys = {'PollPuller_bot': '5131848746:AAEk1LuXl7_0fdN5WzA956t_jjo8Pn6cbl8',
                         'ValeraMainBot': '6225430873:AAEYlbJ2bY-WsLADxlWY1NS-z4r75sf9X5I'}
        self.tb = telebot.TeleBot(self.bot_keys['ValeraMainBot'], False)

        self.chat_ids = {"WTT": -1001179171854, "Alerts": -1001800341082}
        self.default_chat_id = self.chat_ids['Alerts']

    def msg(self, msg, chat_id=None):
        chat_id = self.default_chat_id if chat_id is None else chat_id
        self.tb.send_message(chat_id, msg)

    def img(self, directory, chat_id=None):
        chat_id = self.default_chat_id if chat_id is None else chat_id
        with open(directory, "rb") as img:
            self.tb.send_photo(chat_id, img)


def tg_msg(text=""):
    if text == "":
        caller = get_caller_name()
        text = f"{caller} has finished"
    tb = Telegram()
    tb.msg(text)
# </tg>


def shutdown():
    os.system("shutdown /s /t 1")


def load(path):
    if not path.endswith('.json'):
        path += '.json'
    data = json.load(open(path, 'r'))
    return data


class TimePerfCounters():
    def __init__(self, mark=' * ', autoprint=False):
        self.perfcounters = []
        self.mark = mark
        self.autoprint = autoprint

        self.longest_text = 0
        self.total = 0

    def __call__(self, text=''):
        self.perfcounters.append((time.perf_counter(), text))
        if len(text) > self.longest_text:
            self.longest_text = len(text)

        if self.autoprint and len(self.perfcounters) > 1:
            self.p()

    def p(self, total=True):
        total = self.total + self.perfcounters[-1][0] - self.perfcounters[0][0]
        for i in range(len(self.perfcounters)-1):
            if self.perfcounters[i+1][1] == '':
                continue
            elapsed = self.perfcounters[i+1][0] - self.perfcounters[i][0]
            text = f"{self.perfcounters[i+1][1]:>{self.longest_text}}"
            if self.perfcounters[i+1][1] != '':
                text = f"{text} in: "
            bottleneck_mark = " (!)" if elapsed > total ** 0.75 else ""
            print(f"{self.mark}{text}{elapsed:.2f}{bottleneck_mark}")
        if total:
            empty = f"{' ':>{self.longest_text}}"
            empty += '   : '
            total_print = f"{self.mark}{empty}{total} total"
            print(total_print)
        self.perfcounters = [self.perfcounters[-1]]


def decide_on_datetime_format(*args, format_str='%Y-%m-%d %H:%M:%S'):  # //
    if len(args) == 2:
        CASE = 'strings'
        t1, t2 = args
        if isinstance(t1, datetime):
            dt1, dt2 = t1, t2
        else:
            dt1 = datetime.strptime(t1, format_str)
            dt2 = datetime.strptime(t2, format_str)
    elif len(args) == 1 and isinstance(args[0], list):
        CASE = "list"
        _list = args[0]
        datetime_list = [ts if isinstance(ts, pd.DatetimeIndex) else ts.to_pydatetime(
        ) if isinstance(ts, pd.Timestamp) else None for ts in _list]
        dt1 = next((item for item in datetime_list if item is not None), None)
        dt2 = next((item for item in reversed(
            datetime_list) if item is not None), None)
    elif len(args) == 1 and isinstance(args[0], pd.DatetimeIndex):
        CASE == 'DatetimeIndex'
        print('The hell do I do with a DatetimeIndex?')
    else:
        CASE == 'Unknown'
        raise Exception(
            'Can only process one list or two datetime-like values')

    diff = abs(dt1 - dt2)
    days = diff.days
    hours, remainder = divmod(diff.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    if days > 0:
        new_format_str = '%Y-%m-%d %H:%M:%S'
    elif hours > 0:
        new_format_str = '%H:%M:%S'
    elif minutes > 0:
        new_format_str = '%M:%S'
    else:
        new_format_str = '%S'

    out = CASE
    if CASE == 'strings':
        out = new_format_str
    elif CASE == 'list':
        out = [dt.strftime(new_format_str)
               if dt is not None else None for dt in datetime_list]
    return out


# // move to be a property of .utils.DuckTypes class Timestamp
def round_time_down(dt, round_to=300):
    timestamp = dt.to_ms() / 1000
    rounded_timestamp = timestamp // round_to * round_to
    return datetime.fromtimestamp(rounded_timestamp)


def create_shortcut(source_path, destination_folder):
    import pythoncom
    from win32com.shell import shell

    file_name = os.path.basename(source_path)
    source_folder = os.path.dirname(source_path)
    destination_path = os.path.join(destination_folder, file_name + ".lnk")

    shortcut = pythoncom.CoCreateInstance(
        shell.CLSID_ShellLink,
        None,
        pythoncom.CLSCTX_INPROC_SERVER,
        shell.IID_IShellLink
    )

    shortcut.SetPath(source_path)
    shortcut.SetDescription("Shortcut to " + file_name)
    shortcut.SetIconLocation(source_path, 0)
    shortcut.SetWorkingDirectory(source_folder)

    persist_file = shortcut.QueryInterface(pythoncom.IID_IPersistFile)
    persist_file.Save(os.path.abspath(destination_path), 0)


def time_wrapper(fn, *args, **kwargs):
    shutdown_after = False
    if args[0] == 'shutdown_after':
        shutdown_after = True
        args = args[1:]
    start_time = time.time()
    fn(*args, **kwargs)
    elapsed = time.time() - start_time
    if elapsed > 60:
        tg_msg(f'{fn.__name__} has finished')
    if shutdown_after and elapsed > 180:
        shutdown()


def silent_wrapper(fn, *args, **kwargs):
    try:
        fn(*args, **kwargs)
    except:
        pass


def mkdir(path):
    if not os.path.exists(path):
        os.mkdir(path)


def alert(_input=False):
    dir = os.path.dirname(os.path.abspath(__file__))
    from playsound import playsound
    playsound(os.path.join(dir, 'utils/Notification.mp3'))
    if _input:
        input()


def get_caller_name():
    import inspect
    caller_frame = inspect.stack()[1]
    caller_filename_full = caller_frame.filename
    caller_filename_only = os.path.splitext(
        os.path.basename(caller_filename_full))[0]
    return caller_filename_only


def text_from_image():
    import pytesseract
    from PIL import Image

    pytesseract.pytesseract.tesseract_cmd = r'C:\Program Files\Tesseract-OCR\tesseract.exe'

    t = Image.open("img.png")
    text = pytesseract.image_to_string(t, config='')
    print(text)


if __name__ == '__main__':
    try:
        print('normally I test some function here')
    except Exception as e:
        import traceback
        traceback.print_exc()
    finally:
        from Valera import alert
        alert()



from dataclasses import dataclass
from typing import Generic, TypeVar, Any, ContextManager, Union, NoReturn, Callable, Self
from functools import wraps
from threading import Lock
import contextlib


# Result
# ==============================================================================
T = TypeVar("T")
E = TypeVar("E")
U = TypeVar("U")


def p(func):
	@wraps(func)
	def wrapper(*args, **kwargs):
		try:
			return func(*args, **kwargs)
		except UnwrapOnErr as e:
			raise e
		except UnwrapOnNone as e:
			raise e
		except Exception as e:
			return Err(e)

	return wrapper


class UnwrapOnErr(ValueError):
	def __init__(self, additional_info=None):
		message = "Called `unwrap()` on an `Err` value"
		if additional_info is not None:
			message += f"\n{additional_info}"
		super().__init__(message)


class Result(Generic[T, E]):
	def unwrap(self) -> Union[T, NoReturn]:
		raise NotImplementedError

	def ok_or(self, _or: Any) -> T:
		raise NotImplementedError

	def is_ok(self) -> bool:
		raise NotImplementedError

	def is_err(self) -> bool:
		raise NotImplementedError

	def ok(self):
		raise NotImplementedError

	def or_(self, res: Self) -> Self:
		raise NotImplementedError

	def or_else(self, op: Callable[[E], Self]) -> Self:
		raise NotImplementedError

	def unwrap_or(self, default: T) -> T:
		raise NotImplementedError

	def unwrap_or_else(self, op: Callable[[E], T]) -> T:
		raise NotImplementedError

	def map(self, op: Callable[[T], Self]) -> Self:
		raise NotImplementedError

	def map_err(self, op: Callable[[E], Self]) -> Self:
		raise NotImplementedError

	def p(self) -> Union[T, NoReturn]:
		"""
		Propagate the error or unwrap the value.

		----------------
		To propagate, call .p() on the Result[T, E] and annotate the function with @p. It will run the function in a try-except block and return any errors except for UnwrapOnErr and UnwrapOnNone, defined in this module.
		"""
		raise NotImplementedError


class Ok(Result[T, E]):
	def __init__(self, value: T):
		self.value = value

	def unwrap(self) -> T:
		return self.value

	def ok_or(self, _or: Any) -> T:
		return self.value

	def is_ok(self) -> bool:
		return True

	def is_err(self) -> bool:
		return False

	def ok(self):
		return self.value

	def or_(self, res: Self) -> Self:
		return self

	def or_else(self, op: Callable[[E], Self]) -> Self:
		return self

	def unwrap_or(self, default: T) -> T:
		return self.value

	def unwrap_or_else(self, op: Callable[[E], T]) -> T:
		return self.value

	def map(self, op: Callable[[T], Self]) -> Self:
		return Ok(op(self.value))

	def map_err(self, op: Callable[[E], Self]) -> Self:
		return self

	def p(self) -> T:
		return self.value

	def __str__(self) -> str:
		return f"Ok[{self.value}]"


class Err(Result[T, E]):
	def __init__(self, error: E):
		self.error = error

	def unwrap(self):
		raise UnwrapOnErr(f"{type(self.error).__name__}: {self.error}")  # TODO!: exclude this from traceback

	def ok_or(self, _or: Any) -> Any:
		return _or

	def is_ok(self) -> bool:
		return False

	def is_err(self) -> bool:
		return True

	def ok(self):
		return None

	def or_(self, res: Self) -> Self:
		return res

	def or_else(self, op: Callable[[E], Self]) -> Self:
		return op(self.error)

	def unwrap_or(self, default: T) -> T:
		return default

	def unwrap_or_else(self, op: Callable[[E], T]) -> T:
		return op(self.error)

	def map(self, op: Callable[[T], Self]) -> Self:
		return Err(self.error)

	def map_err(self, op: Callable[[E], Self]) -> Self:
		return Err(op(self.error))

		def p(self) -> NoReturn:
			raise self.error

		def __str__(self) -> str:
			return f'Err[{type(self.error).__name__}["{self.error}"]]'


# Option
# ==============================================================================
class UnwrapOnNone(ValueError):
	def __init__(self, additional_info=None):
		message = "Called `unwrap()` on an `None` value."
		if additional_info is not None:
			message += f"\n{additional_info}"
		super().__init__(message)


class Option(Generic[T]):
	def is_some(self) -> bool:
		raise NotImplementedError

	def is_none(self) -> bool:
		raise NotImplementedError

	def unwrap(self) -> T:
		raise NotImplementedError

	def unwrap_or(self, default: T) -> T:
		raise NotImplementedError

	def map(self, op: Callable[[T], U]) -> "Option[U]":
		raise NotImplementedError

	def map_or(self, default: U, op: Callable[[T], U]) -> U:
		raise NotImplementedError

	def map_or_else(self, default_op: Callable[[], U], op: Callable[[T], U]) -> U:
		raise NotImplementedError

	def ok_or(self, err: E) -> "Result[T, E]":
		raise NotImplementedError

	def ok_or_else(self, err_op: Callable[[], E]) -> "Result[T, E]":
		raise NotImplementedError

	def or_(self, optb: Self) -> Self:
		raise NotImplementedError

	def or_else(self, op: Callable[[], Self]) -> Self:
		raise NotImplementedError

	def replace(self, value: T) -> Self:
		raise NotImplementedError

	def p(self) -> Union[T, NoReturn]:
		"""
		Propagate the error or unwrap the value.

		----------------
		To proagate, call .p() on the Option[object] and annotate the function with @p. It will run the function in a try-except block and return any errors except for UnwrapOnErr and UnwrapOnNone, defined in this module.
		"""
		raise NotImplementedError


class Some(Option[T]):
	def __init__(self, value: T):
		self.value = value

	def is_some(self) -> bool:
		return True

	def is_none(self) -> bool:
		return False

	def unwrap(self) -> T:
		return self.value

	def unwrap_or(self, default: T) -> T:
		return self.value

	def map(self, op: Callable[[T], U]) -> "Option[U]":
		return Some(op(self.value))

	def map_or(self, default: U, op: Callable[[T], U]) -> U:
		return op(self.value)

	def map_or_else(self, default_op: Callable[[], U], op: Callable[[T], U]) -> U:
		return op(self.value)

	def ok_or(self, err: E) -> "Result[T, E]":
		return Ok(self.value)

	def ok_or_else(self, err_op: Callable[[], E]) -> "Result[T, E]":
		return Ok(self.value)

	def or_(self, optb: "Option[T]") -> "Option[T]":
		return self

	def or_else(self, op: Callable[[], "Option[T]"]) -> "Option[T]":
		return self

	def replace(self, value: T) -> "Option[T]":
		old_value = self.value
		self.value = value
		if value is None or value is N:
			self.value = N()
			self.__class__ = N
		return Some(old_value)

	def p(self) -> T:
		return self.value


class N(Option[None]):  # None is named shortly both to avoid conflict with python's None, and to because evokation looks like "N()"; so as to match total number of characters more closely
	def is_some(self) -> bool:
		return False

	def is_none(self) -> bool:
		return True

	def unwrap(self):
		raise UnwrapOnNone("Called `unwrap()` on a `None` value.")

	def unwrap_or(self, default: T) -> T:
		return default

	def map(self, op: Callable[[T], U]) -> "Option[U]":
		return N()

	def map_or(self, default: U, op: Callable[[T], U]) -> U:
		return default

	def map_or_else(self, default_op: Callable[[], U], op: Callable[[T], U]) -> U:
		return default_op()

	def ok_or(self, err: E) -> "Result[T, E]":
		return Err(err)

	def ok_or_else(self, err_op: Callable[[], E]) -> "Result[T, E]":
		return Err(err_op())

	def or_(self, optb: "Option[T]") -> "Option[T]":
		return optb

	def or_else(self, op: Callable[[], "Option[T]"]) -> "Option[T]":
		return op()

	def replace(self, value: T) -> "Option[T]":
		old_value = N()
		if value is not None and value is not N:
			self.__class__ = Some
			self.value = value
		return old_value

	def p(self) -> NoReturn:
		raise RuntimeError("Called `unwrap()` on an `None` value.")


# Mutex
# ==============================================================================
@dataclass
class Mutex(Generic[T]):
	__value: T
	__lock: Lock
	__dropped: bool = False
	"""
		Rust-like Mutex with lock guarantees

		----------------
		Usage:
		```python
			# Create a mutex wrapping the data
		mutex = Mutex([])

	# Lock the mutex for the scope of the `with` block
		with mutex.lock() as value:
			# value is typed as `list` here
			value.append(1)
		```
	"""

	def __init__(self, value: T):
		# Name it with two underscores to make it a bit harder to accidentally access the value from the outside.
		self.__value = value
		self.__lock = Lock()

	@contextlib.contextmanager
	def lock(self) -> Result[ContextManager[T], RuntimeError]:
		if self.__dropped:
			yield Err(RuntimeError("Mutex has been dropped"))
		self.__lock.acquire()
		try:
			yield Ok(self.__value)
		finally:
			self.__lock.release()

	def __str__(self) -> str:
		return f"Mutex[{self.__value}]"

	def take(self) -> Result[T, RuntimeError]:
		if self.__dropped:
			return Err(RuntimeError("Mutex has been dropped"))
		self.__dropped = True
		return Ok(self.__value)



# the default general-purpose package (try not to overload it)
from .Valera import *
from .Binance import Binance as Binance
from .utils.DuckTypes import *
from .utils.rust_types import (
	Ok,
	Err,
	Result,
	UnwrapOnErr,
	Option,
	Some,
	N,
	UnwrapOnNone,
	p,
)

chk = TimePerfCounters()
dtf = decide_on_datetime_format
rtd = round_time_down
tw = time_wrapper
sw = silent_wrapper



import os, json, requests, threading, pandas as pd, numpy as np, time
from datetime import datetime, timedelta

from .utils.DuckTypes import *

class Binance():
    def __init__(self):
        self.PerpSymbols:list[str] = getSymbols('perp')
        self.SpotSymbols:list[str] = getSymbols('spot')
        self.Symbols:list[str] = list(set(self.PerpSymbols+self.SpotSymbols))
       
    #TODO! rolling. preemptively calculate the needed queries, request all at once, then merge.
    def GetKlines(self, symbol:USDTSymbol, startTime:Timestamp, endTime:Timestamp, tf:BinanceTf, market:Market) -> Klines:
        symbol, startTimestamp, endTimestamp, tf, market = USDTSymbol(symbol), Timestamp(startTime), Timestamp(endTime), BinanceTf(tf), Market(market)
        assert startTimestamp.Ms < endTimestamp.Ms, "Your dumb ass has likely switched up endTime and startTime in arguments"
        assert symbol.V in self.Symbols, f"Unknown symbol: {symbol.V}"
        
        params = {
            "symbol": symbol.V,
            "interval": tf.V,
            "startTime": startTimestamp.Ms,
            "endTime": endTimestamp.Ms,
            # this is automatically scaled down when needed, but never up. If not stated, always will cap at 500.
            "limit": 1000,
        }
        url = market.BinanceBaseUrl + '/klines'

        raw_data = requests.get(url, params=params).json()
        df = pd.DataFrame(raw_data, columns=['open_ms', 'open', 'high', 'low', 'close', 'volume', 'close_ms', 'quote_asset_volume', 'trades', 'taker_buy_base', 'taker_buy_quote', 'ignore'])
        
        df['close_dt'] = pd.to_datetime(df['close_ms'], unit='ms')
        df.set_index('close_dt', inplace=True, drop=False)
        df = df[['open', 'high', 'low', 'close', 'quote_asset_volume', 'trades']]
        df = df.rename(columns={"quote_asset_volume":"volume"})

        k = Klines(df)
        k.Market = market.V
        k.Tf = tf.V
        return k

    def CollectKlinesForSymbols(self, symbols:list[USDTSymbol], startTime:Timestamp, endTime:Timestamp, tf:BinanceTf, market:Market=None, full:bool=False) -> SymbolsKlines:
        """
        Takes 15-20s
        """
        if symbols == 'perp':
            symbols = self.PerpSymbols
        if symbols == 'spot':
            symbols = self.SpotSymbols
        if symbols == self.PerpSymbols and market==None:
            market = 'perp'
        if symbols == self.SpotSymbols and market==None:
            market = 'spot'
        assert market != None, "Provide `market:Market` argument too"
        _, _, _, _, _ = [USDTSymbol(s) for s in symbols], Timestamp(startTime), Timestamp(endTime), BinanceTf(tf), Market(market)
        failedOn_buffer = []
        errors = []
        def fetch_data(symbol, klinesDict):
            try:
                k = self.GetKlines(symbol, startTime, endTime, tf, market)
                if full:
                    k = self.GetFullData(k)
                klinesDict[symbol] = k
            except Exception as e:
                failedOn_buffer.append(symbol)
                errors.append(f"{type(e).__name__}: {str(e)}")
        klinesDict = {}
        threads = []
        for symbol in symbols:
            thread = threading.Thread(target=fetch_data, args=(symbol, klinesDict))
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()
        
        if len(failedOn_buffer) != 0:
            print(f"**DEBUG**: Failed to fetch data for {len(failedOn_buffer)} symbols: {failedOn_buffer}")
        assert len(failedOn_buffer)<len(symbols)/2, f"Request failed. Errors encountered: {list(set(errors))}"
        sk = SymbolsKlines(klinesDict)
        return sk
    
    def GetFullData(k:Klines):
        k = Klines(k)
        k = loadNormVolume(k)
        return k
        
        
    def Dump(self, market: Market):
        #todo make it dump the object itself
        _ = Market(market)
        symbols = self.PerpSymbols if market=="perp" else self.SpotSymbols
        
        import inspect
        caller_frame = inspect.stack()[1]
        caller_filename_full = caller_frame.filename
        file_name = f'binance-{market}-pairs.json'
        caller_dir = os.path.dirname(caller_filename_full)
        if 'src' in os.listdir(caller_dir):
            file_name = 'src/' + file_name
        json.dump(symbols, open(file_name, 'w'))
        
def getSymbols(market: Market) -> list[str]:
    market = Market(market)
    """
    returns full list of USDT tickers from the specified market; f: ['BTCUSDT', ...]
    """
    
    url = market.BinanceBaseUrl + '/ticker/24hr'
    to_drop = ['USDC', 'BTCDOM', 'BTCST']
    
    r = requests.get(url)
    assert r.status_code == requests.codes.ok, f"get_perp_symbors()'s request failed with status code: {r.status_code}"
    data = r.json()
    symbols = [ticker['symbol'] for ticker in data if ticker['symbol'][-4:]=='USDT' and not ticker['symbol'][:-4] in to_drop]
    return symbols

def avVolumeCall(ticker, endTime, debug):
    if debug:
        from Valera import d
    base_url = 'https://fapi.binance.com/fapi/v1'
    params = {
        'symbol': ticker,
        'interval': '1h',
        'endTime': endTime,
        'limit': 30*24,
    }
    r = requests.get(base_url+'/klines', params=params).json()

    av_1s_volume = sum([float(item[7]) for item in r])/(30*24*60*60)
    print(f"DEBUG: av 1s volume: {av_1s_volume}") if debug else None
    return av_1s_volume
def loadNormVolume(k:Klines):
    k = Klines(k)
    m, _ = Market(k.Market), BinanceTf(k.Tf)
    baseUrl = m.BinanceBaseUrl
    bounds = (Timestamp(k.V.index[0]), Timestamp(k.V.index[-1]))
    
    params = {
        'symbol': k.Ticker,
        'interval': None,
        'endTime': None,
        'limit': 30*24,
    }
    
    # load 30d
    tf = BinanceTf('1h')
    timeBias_ms = 2*tf.Seconds*1000
    params['interval'], params['endTime'] = tf.V, bounds[0].Ms - timeBias_ms
    
    
    # load hours
    # load last 5m
    return k


if __name__=='__main__':
    try:
        b = Binance()
        out = b.CollectKlinesForSymbols(b.PerpSymbols, time.time()-60*24*24, time.time(), '5m')
        from pprint import pprint
        pprint(out)
    except Exception as e:
        import traceback
        traceback.print_exc()
    finally:
        from Valera import alert
        alert()



import pytest
from ValeraLib import Option, Some, N, UnwrapOnNone, Ok, Err, Result, UnwrapOnErr, p  # noqa: F401


# Test for Some
def test_some_is_some():
	assert Some(5).is_some() is True


def test_some_is_none():
	assert Some(5).is_none() is False


def test_some_unwrap():
	assert Some(5).unwrap() == 5


def test_some_unwrap_or():
	assert Some(5).unwrap_or(10) == 5


def test_some_map():
	assert Some(5).map(lambda x: x * 2).unwrap() == 10


def test_some_map_or():
	assert Some(5).map_or(0, lambda x: x * 2) == 10


def test_some_map_or_else():
	assert Some(5).map_or_else(lambda: 0, lambda x: x * 2) == 10


def test_some_ok_or():
	assert isinstance(Some(5).ok_or("Error"), Ok)
	assert Some(5).ok_or("Error").unwrap() == 5


def test_some_ok_or_else():
	assert isinstance(Some(5).ok_or_else(lambda: "Error"), Ok)
	assert Some(5).ok_or_else(lambda: "Error").unwrap() == 5


def test_some_or():
	assert Some(5).or_(Some(10)).unwrap() == 5


def test_some_or_else():
	assert Some(5).or_else(lambda: Some(10)).unwrap() == 5


def test_some_replace():
	option = Some(5)
	old_value = option.replace(10)
	assert old_value.unwrap() == 5
	assert option.unwrap() == 10


def test_some_p():
	assert Some(5).p() == 5


# Test for N
def test_none_is_some():
	assert N().is_some() is False


def test_none_is_none():
	assert N().is_none() is True


def test_none_unwrap():
	with pytest.raises(UnwrapOnNone):
		N().unwrap()


def test_none_unwrap_or():
	assert N().unwrap_or(10) == 10


def test_none_map():
	assert isinstance(N().map(lambda x: x * 2), N)


def test_none_map_or():
	assert N().map_or(0, lambda x: x * 2) == 0


def test_none_map_or_else():
	assert N().map_or_else(lambda: 0, lambda x: x * 2) == 0


def test_none_ok_or():
	assert isinstance(N().ok_or("Error"), Err)
	N().ok_or("Error") == Err("Error")


def test_none_ok_or_else():
	assert isinstance(N().ok_or_else(lambda: "Error"), Err)
	with pytest.raises(UnwrapOnErr):
		N().ok_or_else(lambda: "Error").unwrap()


def test_none_or():
	assert N().or_(Some(10)).unwrap() == 10


def test_none_or_else():
	assert N().or_else(lambda: Some(10)).unwrap() == 10


def test_none_replace():
	x = N()
	old = x.replace(10)
	assert isinstance(old, N)
	assert x.unwrap() == 10


def test_none_p():
	@p
	def f():
		N().p()

	with pytest.raises(UnwrapOnErr):
		f().unwrap()



import pytest
from ValeraLib import Result, Ok, Err, UnwrapOnErr  # noqa: F401


# Test for Ok
def test_ok_unwrap():
	assert Ok(5).unwrap() == 5


def test_ok_is_ok():
	assert Ok(5).is_ok() is True


def test_ok_is_err():
	assert Ok(5).is_err() is False


def test_ok_map():
	assert Ok(5).map(lambda x: x * 2).unwrap() == 10


def test_ok_map_err():
	assert Ok(5).map_err(lambda x: x * 2).unwrap() == 5  # Error mapper should not be called


def test_ok_or_else():
	result = Ok(5).or_else(lambda x: Err(x))
	assert isinstance(result, Ok)
	assert result.unwrap() == 5


def test_ok_unwrap_or():
	assert Ok(5).unwrap_or(10) == 5


def test_ok_unwrap_or_else():
	assert Ok(2).unwrap_or_else(lambda e: len(e)) == 2


# Test for Err
def test_err_unwrap():
	with pytest.raises(UnwrapOnErr):
		Err("error").unwrap()


def test_err_is_ok():
	assert Err("error").is_ok() is False


def test_err_is_err():
	assert Err("error").is_err() is True


def test_err_map():
	result = Err("error").map(lambda x: x * 2)
	assert isinstance(result, Err)
	with pytest.raises(UnwrapOnErr):
		result.unwrap()


def test_err_map_err():
	result = Err("error").map_err(lambda e: f"new {e}")
	assert isinstance(result, Err)
	with pytest.raises(UnwrapOnErr) as exc_info:
		result.unwrap()
	assert "new error" in str(exc_info.value)


def test_err_or_else():
	result = Err("error").or_else(lambda e: Ok(f"handled {e}"))
	assert isinstance(result, Ok)
	assert result.unwrap() == "handled error"


def test_err_unwrap_or():
	assert Err("error").unwrap_or(10) == 10


def test_err_unwrap_or_else():
	assert Err("foo").unwrap_or_else(lambda e: len(e)) == 3



import plotly.graph_objects as go, numpy as np, pandas as pd, time
from IPython.display import Image

from .utils.DuckTypes import *

def fromRaw(pandas_df):
    """
    do taking log, initializing at 1 and the rest here
    """
    df = pandas_df
    return df

def plotly_closes(normalizedClosesDf:ClosesDf) -> go.Figure:
    #todo move the stats into a separate block, then call this block on closes, and then pass the stats as a dict here, to then plot normally
    _ = ClosesDf(normalizedClosesDf)

    performance = normalizedClosesDf.iloc[-1] - normalizedClosesDf.iloc[0]
    top_performers = performance.nlargest(5).index
    bottom_performers = performance.nsmallest(5).index
    
    mean_values = normalizedClosesDf.mean(axis=1)
    deviations_df = normalizedClosesDf.sub(mean_values, axis=0)
    flattened_deviations = deviations_df.values.flatten()
    variance = np.var(flattened_deviations, ddof=1) #* works, but unconvinient sizing - too many .00s
    kurtosis = pd.Series(flattened_deviations).kurt() #! meaningless due to **4 thingie, that gets skewed immediately on any outliers
    
    correlation_matrix = normalizedClosesDf.corr()
    mean_correlation = correlation_matrix.mean().mean()  #! pretty sure this is a very shitty method
    
    av_move = np.sum(performance)/len(performance)
    # </data-analysis>

    fig = go.Figure()

    def add_trace(*args):
        y, name, line, legend = args
        fig.add_trace(
                go.Scatter(
                    x=normalizedClosesDf.index,
                    y=y,
                    mode='lines',
                    name=name,
                    line=line,
                    showlegend=legend
                )
            )
    def add_performers(column):
        symbol = column[:-4]
        symbol.replace('1000', '', 1)
        sign = f"{performance[column]:+}"[0]
        change = f"{round(100*performance[column], 2):.2f}"
        change = change[1:] if change[0]=='-' else change
        name = f"{symbol:<5}{sign}{change:>5}%"
        add_trace(normalizedClosesDf[column], name, dict(width=2), True)
    def add_empty(name):
        add_trace([1]*len(normalizedClosesDf.index), name, dict(width=0), True)

    # <plotting>
    for column in normalizedClosesDf.columns:
        if column not in top_performers and column not in bottom_performers and column != 'BTCUSDT':
            add_trace(normalizedClosesDf[column], '', dict(width=1, color='grey'), False)
    for column in top_performers:
        add_performers(column)
    add_trace(normalizedClosesDf['BTCUSDT'], f"~BTC~ {round(100*performance['BTCUSDT'], 2):>5}", dict(width=5, color='gold'), True)
    for column in bottom_performers[::-1]:
        add_performers(column)
    add_empty('')
    add_empty(f"V:  {variance:.5f}")
    add_empty(f"K:  {round(kurtosis, 1)}")
    add_empty(f"C:  {round(mean_correlation, 2)}")
    add_empty(f"AV: {av_move*100:.5f}%")
    # </plotting>
    
    fig.update_layout(template='plotly_dark', autosize=True, margin=dict(l=0, r=0, b=0, t=0), font={"family":"Courier New, monospace"})
    fig.update_xaxes(range=[normalizedClosesDf.index.min(), normalizedClosesDf.index.max()])
    fig.update_yaxes(range=[normalizedClosesDf.min().min(), normalizedClosesDf.max().max()])

    return fig

if __name__=='__main__':
    try:
        from Binance import Binance
        b = Binance()
        data = b.CollectKlinesForSymbols('perp', time.time()-24*60*60, time.time(), '5m')
        plotly_closes(data)
    except Exception as e:
        import traceback
        traceback.print_exc()
    finally:
        from Valera import alert
        alert()